(declare (block) (standard-bindings) (extended-bindings))
(begin
  (define-macro (define-guard guard defn)
    (if (eval `(cond-expand (,guard #t) (else #f)))
        '(begin)
        (begin (eval `(define-cond-expand-feature ,guard)) defn)))
  (define-macro (define-c-lambda id args ret #!optional (name #f))
    (let ((name (or name (symbol->string id))))
      `(define ,id (c-lambda ,args ,ret ,name))))
  (define-macro (define-const symbol)
    (let* ((str (symbol->string symbol))
           (ref (string-append "___return (" str ");")))
      `(define ,symbol ((c-lambda () int ,ref)))))
  (define-macro (define-const* symbol #!optional (ccond #f))
    (let* ((str (symbol->string symbol))
           (code (string-append
                  "#if "
                  (or ccond (string-append "defined(" str ")"))
                  "\n"
                  "___return (___FIX ("
                  str
                  "));\n"
                  "#else \n"
                  "___return (___FAL);\n"
                  "#endif")))
      `(define ,symbol ((c-lambda () scheme-object ,code)))))
  (define-macro (define-with-errno symbol ffi-symbol args)
    `(define (,symbol ,@args)
       (declare (not interrupts-enabled))
       (let ((r (,ffi-symbol ,@args)))
         (if (##fx< r 0) (##fx- (##c-code "___RESULT = ___FIX (errno);")) r))))
  (define-macro (define-c-struct
                 struct
                 #!optional
                 (members '())
                 release-function)
    (let* ((struct-str (symbol->string struct))
           (struct-ptr (string->symbol (string-append struct-str "*")))
           (shallow-ptr
            (string->symbol (string-append struct-str "-shallow-ptr*")))
           (borrowed-ptr
            (string->symbol (string-append struct-str "-borrowed-ptr*")))
           (string-types
            '(char-string
              nonull-char-string
              UTF-8-string
              nonnull-UTF-8-string
              UTF-16-string
              nonnull-UTF16-string))
           (string-compat-required?
            (let loop ((m members))
              (cond ((null? m) #f)
                    ((member (cdr (car m)) string-types) #t)
                    (else (loop (cdr m))))))
           (string-setter-body
            (lambda (member-name)
              (let ((m (string-append "___arg1->" member-name)))
                (string-append
                 "if("
                 m
                 " == NULL)"
                 "\n"
                 m
                 "= strdup(___arg2);"
                 "\n"
                 "else if (strcmp("
                 m
                 ", ___arg2) != 0) {"
                 "\n"
                 "free("
                 m
                 ");"
                 "\n"
                 m
                 "= strdup(___arg2);"
                 "\n"
                 "}"
                 "\n"
                 "___return;"
                 "\n"))))
           (default-free-body
            (and string-compat-required?
                 (string-append
                  "___SCMOBJ "
                  struct-str
                  "_ffi_free (void *ptr) {"
                  "\n"
                  "struct "
                  struct-str
                  " *obj = (struct "
                  struct-str
                  "*) ptr;"
                  "\n"
                  (apply string-append
                         (map (lambda (m)
                                (cond ((memq (cdr m) string-types)
                                       (let ((mem-name
                                              (symbol->string (car m))))
                                         (string-append
                                          "if(obj->"
                                          mem-name
                                          ") "
                                          "free(obj->"
                                          mem-name
                                          ");"
                                          "\n")))
                                      (else "")))
                              members))
                  "free(obj);"
                  "\n"
                  "return ___FIX (___NO_ERR);"
                  "\n"
                  "}")))
           (release-function
            (or release-function
                (if string-compat-required?
                    (string-append struct-str "_ffi_free")
                    "ffi_free")))
           (string-compat-types
            (if string-compat-required?
                `((c-declare ,default-free-body)
                  (c-define-type
                   ,shallow-ptr
                   (pointer ,struct (,struct-ptr) "ffi_free")))
                '())))
      `(begin
         (c-define-type ,struct (struct ,struct-str))
         (c-define-type
          ,struct-ptr
          (pointer ,struct (,struct-ptr) ,release-function))
         (c-define-type ,borrowed-ptr (pointer ,struct (,struct-ptr)))
         ,@string-compat-types
         (define ,(string->symbol (string-append struct-str "-ptr?"))
           (lambda (obj)
             (and (foreign? obj) (equal? (foreign-tags obj) '(,struct-ptr)))))
         ,@(apply append
                  (map (lambda (m)
                         (let* ((member-name (symbol->string (car m)))
                                (member-type (cdr m))
                                (getter-name
                                 (string-append struct-str "-" member-name))
                                (setter-body
                                 (cond ((member member-type string-types)
                                        (string-setter-body member-name))
                                       (else
                                        (string-append
                                         "___arg1->"
                                         member-name
                                         " = ___arg2;"
                                         "\n"
                                         "___return;"
                                         "\n")))))
                           `((define ,(string->symbol getter-name)
                               (c-lambda
                                (,struct-ptr)
                                ,member-type
                                ,(string-append
                                  "___return(___arg1->"
                                  member-name
                                  ");")))
                             (define ,(string->symbol
                                       (string-append getter-name "-set!"))
                               (c-lambda
                                (,struct-ptr ,member-type)
                                void
                                ,setter-body)))))
                       members))
         (define ,(string->symbol (string-append "malloc-" struct-str))
           (c-lambda
            ()
            ,struct-ptr
            ,(string-append
              "struct "
              struct-str
              "* var = (struct "
              struct-str
              " *) malloc(sizeof(struct "
              struct-str
              "));"
              "\n"
              "if (var == NULL)"
              "\n"
              "    ___return (NULL);"
              "\n"
              "memset(var, 0, sizeof(struct "
              struct-str
              "));"
              "___return(var);")))
         (define ,(string->symbol (string-append "ptr->" struct-str))
           (c-lambda (,struct-ptr) ,struct "___return(*___arg1);"))
         (define ,(string->symbol
                   (string-append "malloc-" struct-str "-array"))
           (c-lambda
            (unsigned-int32)
            ,(if string-compat-required? shallow-ptr struct-ptr)
            ,(string-append
              "struct "
              struct-str
              " *arr_var=(struct "
              struct-str
              " *) malloc(___arg1*sizeof(struct "
              struct-str
              "));"
              "\n"
              "if (arr_var == NULL)"
              "\n"
              "    ___return (NULL);"
              "\n"
              "memset(arr_var, 0, ___arg1*sizeof(struct "
              struct-str
              "));"
              "\n"
              "___return(arr_var);")))
         (define ,(string->symbol (string-append struct-str "-array-ref"))
           (c-lambda
            (,struct-ptr unsigned-int32)
            ,borrowed-ptr
            "___return (___arg1 + ___arg2);"))
         (define ,(string->symbol (string-append struct-str "-array-set!"))
           (c-lambda
            (,struct-ptr unsigned-int32 ,struct-ptr)
            void
            "*(___arg1 + ___arg2) = *___arg3; ___return;")))))
  (c-declare "#include <stdlib.h>")
  (c-declare "#include <string.h>")
  (c-declare "#include <errno.h>")
  (c-declare "static ___SCMOBJ ffi_free (void *ptr);")
  (c-declare
   "#ifndef ___HAVE_FFI_U8VECTOR\n#define ___HAVE_FFI_U8VECTOR\n#define U8_DATA(obj) ___CAST (___U8*, ___BODY_AS (obj, ___tSUBTYPED))\n#define U8_LEN(obj) ___HD_BYTES (___HEADER (obj))\n#endif")
  (namespace
   ("std/foreign-test#"
    g
    bar-j-set!
    bar-j
    bar-i-set!
    bar-i
    bar-ptr?
    ptr->bar
    bar-array-set!
    bar-array-ref
    malloc-bar-array
    malloc-bar
    foo-str-set!
    foo-str
    foo-d2-set!
    foo-d2
    foo-a1-set!
    foo-a1
    foo-ptr?
    ptr->foo
    foo-array-set!
    foo-array-ref
    malloc-foo-array
    malloc-foo
    abc-b-set!
    abc-b
    abc-a-set!
    abc-a
    abc-ptr?
    ptr->abc
    abc-array-set!
    abc-array-ref
    malloc-abc-array
    malloc-abc
    f))
  (c-declare
   "\nstruct abc {\n    char* a;\n    char* b;\n    char* c;\n};\nstruct d { int e;};\n\nstruct foo {\n    struct abc* a1;\n    struct abc* d2;\n    char* str;\n};\n\nstruct bar {\nint i;\nint j;\n};\n")
  (c-define-type d (struct "d"))
  (define-c-lambda f () int "___return(2);")
  (define-c-lambda g () int "___return(2);")
  (define-c-struct abc ((a . char-string) (b . char-string)))
  (define-c-struct foo ((a1 . abc*) (d2 . abc*) (str . char-string)))
  (define-c-struct bar ((i . int) (j . int)))
  (c-declare
   "#ifndef ___HAVE_FFI_FREE\n#define ___HAVE_FFI_FREE\n___SCMOBJ ffi_free (void *ptr)\n{\n free (ptr);\n return ___FIX (___NO_ERR);\n}\n#endif")
  (define std/foreign-test#foreign-test
    (std/test#make-test-suite
     '"test :std/foreign"
     (lambda ()
       (letrec* ((_test-str145077_ '"hello")
                 (_test-str245078_ '"world")
                 (_make-abc45079_
                  (lambda (_a45148_ _b45149_)
                    (let ((_o45151_ (std/foreign-test#malloc-abc)))
                      (std/foreign-test#abc-a-set! _o45151_ _a45148_)
                      (std/foreign-test#abc-b-set! _o45151_ _b45149_)
                      _o45151_))))
         (std/test#run-test-case!
          '"c struct"
          (lambda ()
            (letrec* ((_obj45082_ (std/foreign-test#malloc-abc)))
              (std/foreign-test#abc-a-set! _obj45082_ _test-str145077_)
              (std/foreign-test#abc-b-set! _obj45082_ _test-str245078_)
              (let ((_val45084_ _test-str145077_))
                (std/test#verbose
                 '"... check ~a is ~a to ~s~n"
                 '(abc-a obj)
                 'equal?
                 _val45084_)
                (std/test#test-check-e
                 '(check equal? (abc-a obj) test-str1)
                 std/test#equal-values?
                 (lambda () (std/foreign-test#abc-a _obj45082_))
                 _val45084_
                 '"\"foreign-test.ss\"@62.33"))
              (let ((_val45088_ _test-str245078_))
                (std/test#verbose
                 '"... check ~a is ~a to ~s~n"
                 '(abc-b obj)
                 'equal?
                 _val45088_)
                (std/test#test-check-e
                 '(check equal? (abc-b obj) test-str2)
                 std/test#equal-values?
                 (lambda () (std/foreign-test#abc-b _obj45082_))
                 _val45088_
                 '"\"foreign-test.ss\"@63.33"))
              (let ((_val45092_ '#t))
                (std/test#verbose
                 '"... check ~a is ~a to ~s~n"
                 '(abc-ptr? obj)
                 'equal?
                 _val45092_)
                (std/test#test-check-e
                 '(check equal? (abc-ptr? obj) #t)
                 std/test#equal-values?
                 (lambda () (std/foreign-test#abc-ptr? _obj45082_))
                 _val45092_
                 '"\"foreign-test.ss\"@65.33")))))
         (std/test#run-test-case!
          '"c struct array"
          (lambda ()
            (letrec* ((_obj-array45097_ (std/foreign-test#malloc-abc-array '2))
                      (_obj145098_ (std/foreign-test#malloc-abc))
                      (_g301935_
                       (let ((_g301936_
                              (std/foreign-test#abc-a-set!
                               _obj145098_
                               _test-str145077_)))
                         #!void
                         _g301936_))
                      (_obj245099_ (std/foreign-test#malloc-abc)))
              (begin
                (std/foreign-test#abc-a-set! _obj245099_ _test-str245078_)
                (std/foreign-test#abc-array-set!
                 _obj-array45097_
                 '0
                 _obj145098_)
                (std/foreign-test#abc-array-set!
                 _obj-array45097_
                 '1
                 _obj245099_)
                (let ((_val45101_ _test-str145077_))
                  (std/test#verbose
                   '"... check ~a is ~a to ~s~n"
                   '(abc-a (abc-array-ref obj-array 0))
                   'equal?
                   _val45101_)
                  (std/test#test-check-e
                   '(check equal?
                           (abc-a (abc-array-ref obj-array 0))
                           test-str1)
                   std/test#equal-values?
                   (lambda ()
                     (std/foreign-test#abc-a
                      (std/foreign-test#abc-array-ref _obj-array45097_ '0)))
                   _val45101_
                   '"\"foreign-test.ss\"@80.33"))
                (let ((_val45105_ _test-str245078_))
                  (std/test#verbose
                   '"... check ~a is ~a to ~s~n"
                   '(abc-a (abc-array-ref obj-array 1))
                   'equal?
                   _val45105_)
                  (std/test#test-check-e
                   '(check equal?
                           (abc-a (abc-array-ref obj-array 1))
                           test-str2)
                   std/test#equal-values?
                   (lambda ()
                     (std/foreign-test#abc-a
                      (std/foreign-test#abc-array-ref _obj-array45097_ '1)))
                   _val45105_
                   '"\"foreign-test.ss\"@81.33"))))))
         (std/test#run-test-case!
          '"c int struct array"
          (lambda ()
            (letrec* ((_y45110_ (std/foreign-test#malloc-bar))
                      (_g301938_
                       (let ((_g301939_
                              (std/foreign-test#bar-i-set! _y45110_ '123)))
                         #!void
                         _g301939_))
                      (_g301941_
                       (let ((_g301942_
                              (std/foreign-test#bar-j-set! _y45110_ '456)))
                         #!void
                         _g301942_))
                      (_y245111_ (std/foreign-test#malloc-bar))
                      (_g301944_
                       (let ((_g301945_
                              (std/foreign-test#bar-i-set! _y245111_ '320)))
                         #!void
                         _g301945_))
                      (_g301947_
                       (let ((_g301948_
                              (std/foreign-test#bar-j-set! _y245111_ '328)))
                         #!void
                         _g301948_))
                      (_k45112_ (std/foreign-test#malloc-bar-array '2)))
              (begin
                (std/foreign-test#bar-array-set! _k45112_ '0 _y45110_)
                (std/foreign-test#bar-array-set! _k45112_ '1 _y245111_)
                (let ((_val45114_ '320))
                  (std/test#verbose
                   '"... check ~a is ~a to ~s~n"
                   '(bar-i (bar-array-ref k 1))
                   'equal?
                   _val45114_)
                  (std/test#test-check-e
                   '(check equal? (bar-i (bar-array-ref k 1)) 320)
                   std/test#equal-values?
                   (lambda ()
                     (std/foreign-test#bar-i
                      (std/foreign-test#bar-array-ref _k45112_ '1)))
                   _val45114_
                   '"\"foreign-test.ss\"@96.33"))))))
         (std/test#run-test-case!
          '"modifying c strings"
          (lambda ()
            (letrec* ((_obj145119_ (std/foreign-test#malloc-abc))
                      (_t345120_
                       (string-append _test-str145077_ _test-str245078_))
                      (_t445121_ (string-append _test-str145077_ '" test")))
              (std/foreign-test#abc-a-set! _obj145119_ _test-str145077_)
              (std/foreign-test#abc-b-set! _obj145119_ _test-str245078_)
              (std/foreign-test#abc-a-set! _obj145119_ _t345120_)
              (std/foreign-test#abc-b-set! _obj145119_ _t445121_)
              (let ((_val45123_ _t345120_))
                (std/test#verbose
                 '"... check ~a is ~a to ~s~n"
                 '(abc-a obj1)
                 'equal?
                 _val45123_)
                (std/test#test-check-e
                 '(check equal? (abc-a obj1) t3)
                 std/test#equal-values?
                 (lambda () (std/foreign-test#abc-a _obj145119_))
                 _val45123_
                 '"\"foreign-test.ss\"@109.33"))
              (let ((_val45127_ _t445121_))
                (std/test#verbose
                 '"... check ~a is ~a to ~s~n"
                 '(abc-b obj1)
                 'equal?
                 _val45127_)
                (std/test#test-check-e
                 '(check equal? (abc-b obj1) t4)
                 std/test#equal-values?
                 (lambda () (std/foreign-test#abc-b _obj145119_))
                 _val45127_
                 '"\"foreign-test.ss\"@110.33")))))
         (std/test#run-test-case!
          '"nested structs"
          (lambda ()
            (letrec* ((_foo-arr45132_ (std/foreign-test#malloc-foo-array '2))
                      (_obj145133_
                       (_make-abc45079_ _test-str145077_ _test-str245078_))
                      (_obj245134_
                       (_make-abc45079_ _test-str245078_ _test-str145077_))
                      (_t545135_ '"here we go")
                      (_t645136_ '"but not")
                      (_foo145137_ (std/foreign-test#malloc-foo))
                      (_g301950_
                       (let ((_g301951_
                              (std/foreign-test#foo-a1-set!
                               _foo145137_
                               _obj145133_)))
                         #!void
                         _g301951_))
                      (_g301953_
                       (let ((_g301954_
                              (std/foreign-test#foo-d2-set!
                               _foo145137_
                               _obj245134_)))
                         #!void
                         _g301954_))
                      (_g301956_
                       (let ((_g301957_
                              (std/foreign-test#foo-str-set!
                               _foo145137_
                               _t545135_)))
                         #!void
                         _g301957_))
                      (_foo245138_ (std/foreign-test#malloc-foo)))
              (begin
                (std/foreign-test#foo-a1-set! _foo245138_ _obj245134_)
                (std/foreign-test#foo-d2-set! _foo245138_ _obj145133_)
                (std/foreign-test#foo-str-set! _foo245138_ _t645136_)
                (std/foreign-test#foo-array-set! _foo-arr45132_ '0 _foo145137_)
                (std/foreign-test#foo-array-set! _foo-arr45132_ '1 _foo245138_)
                (let ((_val45140_ _t545135_))
                  (std/test#verbose
                   '"... check ~a is ~a to ~s~n"
                   '(foo-str (foo-array-ref foo-arr 0))
                   'equal?
                   _val45140_)
                  (std/test#test-check-e
                   '(check equal? (foo-str (foo-array-ref foo-arr 0)) t5)
                   std/test#equal-values?
                   (lambda ()
                     (std/foreign-test#foo-str
                      (std/foreign-test#foo-array-ref _foo-arr45132_ '0)))
                   _val45140_
                   '"\"foreign-test.ss\"@134.33"))
                (let ((_val45144_ _t645136_))
                  (std/test#verbose
                   '"... check ~a is ~a to ~s~n"
                   '(foo-str (foo-array-ref foo-arr 1))
                   'equal?
                   _val45144_)
                  (std/test#test-check-e
                   '(check equal? (foo-str (foo-array-ref foo-arr 1)) t6)
                   std/test#equal-values?
                   (lambda ()
                     (std/foreign-test#foo-str
                      (std/foreign-test#foo-array-ref _foo-arr45132_ '1)))
                   _val45144_
                   '"\"foreign-test.ss\"@135.33")))))))))))
